from functools import partial

from torchnlp.encoders.text.static_tokenizer_encoder import StaticTokenizerEncoder


def _tokenize(s, delimiter):
    return s.split(delimiter)


class DelimiterEncoder(StaticTokenizerEncoder):
    """ Encodes text into a tensor by splitting the text using a delimiter.

    Args:
        delimiter (string): Delimiter used with ``string.split``
        sample (list): Sample of data used to build encoding dictionary.
        min_occurrences (int, optional): Minimum number of occurrences for a token to be added to
          the encoding dictionary.
        append_eos (bool, optional): If ``True`` append EOS token onto the end to the encoded
          vector.
        reserved_tokens (list of str, optional): List of reserved tokens inserted in the beginning
            of the dictionary.
        eos_index (int, optional): The eos token is used to encode the end of a sequence. This is
          the index that token resides at.
        unknown_index (int, optional): The unknown token is used to encode unseen tokens. This is
          the index that token resides at.
        padding_index (int, optional): The unknown token is used to encode sequence padding. This is
          the index that token resides at.


    Example:

        >>> encoder = DelimiterEncoder('|', ['token_a|token_b', 'token_c'])
        >>> encoder.encode('token_a|token_c')
        tensor([5, 7])
        >>> encoder.vocab
        ['<pad>', '<unk>', '</s>', '<s>', '<copy>', 'token_a', 'token_b', 'token_c']

    """

    def __init__(self, delimiter, *args, **kwargs):
        if 'tokenize' in kwargs:
            raise TypeError('Encoder does not take keyword argument tokenize.')

        self.delimiter = delimiter
        super().__init__(*args, tokenize=partial(_tokenize, delimiter=self.delimiter), **kwargs)

    def decode(self, tensor):
        return self.delimiter.join([self.itos[index] for index in tensor])



import torch


class Encoder(object):
    """ Base class for a encoder.
    """

    def __init__(self):  # pragma: no cover
        raise NotImplementedError

    def enforce_reversible(self):
        """ Updates the specification of ``Encoder.encode`` and ``Encoder.decode`` to enforce
        reversibility.

        Formally, reversible means: ``Encoder.decode(Encoder.encode(object_)) == object_``

        Example:
            >>> encoder = Encoder().enforce_reversible()  # doctest: +SKIP
        """
        last_encode = self.encode
        last_decode = self.decode

        def _encode(object_, *args, **kwargs):
            encoded = last_encode(object_, *args, **kwargs)
            if last_decode(encoded) != object_:
                raise ValueError('Encoding is not reversible for "%s"', object_)
            return encoded

        self.encode = _encode

        def _decode(tensor, *args, **kwargs):
            decoded = last_decode(tensor, *args, **kwargs)
            if torch.equals(last_encode(decoded), tensor):
                raise ValueError('Decode is not reversible for "%s"', tensor)
            return decoded

        self.decode = _decode
        return self

    def encode(self, object_):  # pragma: no cover
        """ Encodes an object to a :class:`torch.Tensor`.

        Args:
            object_ (object): Object to encode.

        Returns:
            torch.Tensor: Encoding of the object.
        """
        raise NotImplementedError

    def batch_encode(self, batch, *args, **kwargs):
        """
        Args:
            batch (list): Batch of objects to encode.
            *args: Arguments passed to ``encode``.
            **kwargs: Key word arguments passed to ``encode``.

        Returns:
            list: Batch of encoded objects.
        """
        return torch.tensor([self.encode(object_, *args, **kwargs).tolist() for object_ in batch])

    def decode(self, tensor):  # pragma: no cover
        """ Decodes a :class:`torch.Tensor` to a object.

        Args:
            tensor (torch.Tensor): Tensor to decode.

        Returns:
            object: Object decoded from tensor.
        """
        raise NotImplementedError

    def batch_decode(self, batch, *args, **kwargs):
        """
        Args:
            batch (list of :class:`torch.Tensor`): Batch of encoded objects.
            *args: Arguments passed to ``decode``.
            **kwargs: Key word arguments passed to ``decode``.

        Returns:
            list: Batch of decoded objects.
        """
        iterator = ([split.squeeze(0) for split in batch.split(1)]
                    if torch.is_tensor(batch) else batch)
        return [self.decode(tensor, *args, **kwargs) for tensor in iterator]



from torchnlp.encoders.text.delimiter_encoder import DelimiterEncoder


class WhitespaceEncoder(DelimiterEncoder):
    """ Encodes a text by splitting on whitespace.

    Args:
        sample (list): Sample of data used to build encoding dictionary.
        min_occurrences (int, optional): Minimum number of occurrences for a token to be added to
          the encoding dictionary.
        append_eos (bool, optional): If ``True`` append EOS token onto the end to the encoded
          vector.
        reserved_tokens (list of str, optional): List of reserved tokens inserted in the beginning
            of the dictionary.
        eos_index (int, optional): The eos token is used to encode the end of a sequence. This is
          the index that token resides at.
        unknown_index (int, optional): The unknown token is used to encode unseen tokens. This is
          the index that token resides at.
        padding_index (int, optional): The unknown token is used to encode sequence padding. This is
          the index that token resides at.

    Example:

        >>> encoder = WhitespaceEncoder(["This ain't funny.", "Don't?"])
        >>> encoder.encode("This ain't funny.")
        tensor([5, 6, 7])
        >>> encoder.vocab
        ['<pad>', '<unk>', '</s>', '<s>', '<copy>', 'This', "ain't", 'funny.', "Don't?"]
        >>> encoder.decode(encoder.encode("This ain't funny."))
        "This ain't funny."

    """

    def __init__(self, *args, **kwargs):
        super().__init__(' ', *args, **kwargs)



from torchnlp.encoders.text.static_tokenizer_encoder import StaticTokenizerEncoder


def _tokenize(s):
    return list(s)


class CharacterEncoder(StaticTokenizerEncoder):
    """ Encodes text into a tensor by splitting the text into individual characters.

    Args:
        sample (list): Sample of data used to build encoding dictionary.
        min_occurrences (int, optional): Minimum number of occurrences for a token to be added to
          the encoding dictionary.
        append_eos (bool, optional): If ``True`` append EOS token onto the end to the encoded
          vector.
        reserved_tokens (list of str, optional): List of reserved tokens inserted in the beginning
            of the dictionary.
        eos_index (int, optional): The eos token is used to encode the end of a sequence. This is
          the index that token resides at.
        unknown_index (int, optional): The unknown token is used to encode unseen tokens. This is
          the index that token resides at.
        padding_index (int, optional): The unknown token is used to encode sequence padding. This is
          the index that token resides at.
    """

    def __init__(self, *args, **kwargs):
        if 'tokenize' in kwargs:
            raise TypeError('Encoder does not take keyword argument tokenize.')

        super().__init__(*args, tokenize=_tokenize, **kwargs)

    def decode(self, tensor):
        return ''.join([self.itos[index] for index in tensor])



from torchnlp.encoders.text.static_tokenizer_encoder import StaticTokenizerEncoder


class MosesEncoder(StaticTokenizerEncoder):
    """ Encodes the text using the Moses tokenizer.

    **Tokenizer Reference:**
    http://www.nltk.org/_modules/nltk/tokenize/moses.html

    Args:
        sample (list): Sample of data used to build encoding dictionary.
        min_occurrences (int, optional): Minimum number of occurrences for a token to be added to
          the encoding dictionary.
        append_eos (bool, optional): If ``True`` append EOS token onto the end to the encoded
          vector.
        reserved_tokens (list of str, optional): List of reserved tokens inserted in the beginning
            of the dictionary.
        eos_index (int, optional): The eos token is used to encode the end of a sequence. This is
          the index that token resides at.
        unknown_index (int, optional): The unknown token is used to encode unseen tokens. This is
          the index that token resides at.
        padding_index (int, optional): The unknown token is used to encode sequence padding. This is
          the index that token resides at.

    Example:

        >>> encoder = MosesEncoder(["This ain't funny.", "Don't?"])
        >>> encoder.encode("This ain't funny.")
        tensor([5, 6, 7, 8, 9])
        >>> encoder.vocab
        ['<pad>', '<unk>', '</s>', '<s>', '<copy>', 'This', 'ain', '&apos;t', 'funny', '.', \
'Don', '?']
        >>> encoder.decode(encoder.encode("This ain't funny."))
        "This ain't funny."

    """

    def __init__(self, *args, **kwargs):
        if 'tokenize' in kwargs:
            raise TypeError('Encoder does not take keyword argument tokenize.')

        try:
            from sacremoses import MosesTokenizer
            from sacremoses import MosesDetokenizer
        except ImportError:
            print("Please install SacreMoses. "
                  "See the docs at https://github.com/alvations/sacremoses for more information.")
            raise

        self.detokenizer = MosesDetokenizer()

        super().__init__(*args, tokenize=MosesTokenizer().tokenize, **kwargs)

    def decode(self, tensor):
        tokens = [self.itos[index] for index in tensor]
        return self.detokenizer.detokenize(tokens, return_str=True)
